@@ -2,7 +2,7 @@
 
 use crate::{
     defs::{LengthType, SpareMemoryPolicy},
-    errors::CapacityError,
+    errors::{CapacityError, CapacityErrorVal, InsertError, InsertErrorVal},
 };
 use core::{marker::PhantomData, mem, ptr, result::Result, slice};
 
@@ -35,7 +35,7 @@ use core::{marker::PhantomData, mem, ptr, result::Result, slice};
 /// let mut v = ArrayVec::<u64, U8, Uninitialized, 12>::new();
 /// assert_eq!(v.len(), 0);
 /// assert_eq!(v.capacity(), 12);
-/// assert_eq!(v.spare_capacity_len(), 12);
+/// assert_eq!(v.spare_capacity(), 12);
 /// assert_eq!(v, []);
 ///
 /// v.push(1);
@@ -43,7 +43,7 @@ use core::{marker::PhantomData, mem, ptr, result::Result, slice};
 ///
 /// assert_eq!(v.len(), 2);
 /// assert_eq!(v.capacity(), 12);
-/// assert_eq!(v.spare_capacity_len(), 10);
+/// assert_eq!(v.spare_capacity(), 10);
 /// assert_eq!(v, [1, 2]);
 ///
 /// v[0] = 7;
@@ -279,7 +279,7 @@ where
         Self::CAPACITY
     }
 
-    /// Returns the number of elements left until the array-vector is completely full.
+    /// Returns the number of elements the array-vector can hold in addition to already held ones.
     ///
     /// Equivalent to `capacity() - len()`.
     ///
@@ -289,13 +289,14 @@ where
     /// # use cds::array_vec;
     /// let mut v = array_vec![2; u64];
     /// assert_eq!(v.capacity(), 2);
-    /// assert_eq!(v.spare_capacity_len(), 2);
+    /// assert_eq!(v.spare_capacity(), 2);
     ///
     /// v.push(1);
-    /// assert_eq!(v.spare_capacity_len(), 1);
+    /// assert_eq!(v.capacity(), 2);
+    /// assert_eq!(v.spare_capacity(), 1);
     /// ```
     #[inline]
-    pub fn spare_capacity_len(&self) -> usize {
+    pub fn spare_capacity(&self) -> usize {
         Self::CAPACITY - self.len.as_usize()
     }
 
@@ -322,6 +323,10 @@ where
 
     /// Forces the length of the array-vector to `new_len`.
     ///
+    /// Note that this method doesn't [`drop`] elements, which may lead to a resource leak if
+    /// `new_len < len()` and `T` has a custom [`Drop`] implementation. See [`truncate`] for a
+    /// method that handles array-vector truncation properly.
+    ///
     /// # Safety
     ///
     /// - `new_len` must be less than or equal to the array-vector's [`CAPACITY`]
@@ -332,12 +337,49 @@ where
     /// # Panics
     ///
     /// This method uses debug assertions to verify that `new_len` is in bounds.
+    ///
+    /// [`drop`]: core::mem::drop
+    /// [`Drop`]: core::ops::Drop
+    /// [`truncate`]: ArrayVec::truncate
     #[inline]
     pub unsafe fn set_len(&mut self, new_len: usize) {
         debug_assert!(new_len <= Self::CAPACITY);
         self.len.set(new_len);
     }
 
+    /// Appends an element to the back of the array-vector.
+    ///
+    /// # Panics
+    ///
+    /// This method panics if there is no spare capacity to accommodate the new element.
+    /// See [`try_push`] for a method that returns an error instead.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// # use cds::array_vec;
+    /// let mut v = array_vec![2; u64];
+    /// v.push(1);
+    /// v.push(2);
+    /// ```
+    ///
+    /// [`push`] panics if there is no spare capacity:
+    ///
+    /// ```should_panic
+    /// # use cds::array_vec;
+    /// let mut v = array_vec![2; u64];
+    /// v.push(1);
+    /// v.push(2);
+    /// v.push(3);  // <-- this panics
+    /// ```
+    ///
+    /// [`try_push`]: ArrayVec::try_push
+    /// [`push`]: ArrayVec::push
+    #[inline]
+    pub fn push(&mut self, e: T) {
+        self.try_push(e).expect("insufficient capacity")
+    }
+
     /// Tries to append an element to the back of the array-vector.
     ///
     /// Returns [`CapacityError`] if there is no spare capacity to accommodate a new element.
@@ -366,37 +408,64 @@ where
         }
     }
 
-    /// Appends an element to the back of the array-vector.
+    /// Tries to append an element to the back of the array-vector.
     ///
-    /// # Panics
+    /// Returns [`CapacityErrorVal`] if there is no spare capacity to accommodate a new element.
     ///
-    /// This method panics of there is no spare capacity to accommodate the new element.
-    /// See [`try_push`] for a method that returns an error instead.
+    /// The difference between this method and [`try_push`] is that in case of an error
+    /// [`try_push_val`] returns the element back to the caller.
+    ///
+    /// This is a non-panic version of [`push`].
     ///
     /// # Examples
-    ///
     /// ```rust
-    /// # use cds::array_vec;
+    /// # use cds::{array_vec, errors::CapacityErrorVal};
     /// let mut v = array_vec![2; u64];
-    /// v.push(1);
-    /// v.push(2);
-    /// ```
+    /// assert_eq!(v, []);
     ///
-    /// [`push`] panics if there is no spare capacity:
+    /// assert!(v.try_push_val(1).is_ok());
+    /// assert!(v.try_push_val(2).is_ok());
+    /// assert_eq!(v, [1, 2]);
     ///
-    /// ```should_panic
-    /// # use cds::array_vec;
-    /// let mut v = array_vec![2; u64];
-    /// v.push(1);
-    /// v.push(2);
-    /// v.push(3);  // <-- this panics
+    /// assert!(matches!(v.try_push_val(3), Err(CapacityErrorVal(e)) if e == 3));
+    /// assert_eq!(v, [1, 2]);
     /// ```
     ///
+    /// [`try_push_val`]: ArrayVec::try_push_val
     /// [`try_push`]: ArrayVec::try_push
     /// [`push`]: ArrayVec::push
     #[inline]
-    pub fn push(&mut self, e: T) {
-        self.try_push(e).expect("ArrayVec::push failed")
+    pub fn try_push_val(&mut self, value: T) -> Result<(), CapacityErrorVal<T>> {
+        if self.len < Self::CAPACITY {
+            unsafe { self.push_unchecked(value) };
+            Ok(())
+        } else {
+            Err(CapacityErrorVal(value))
+        }
+    }
+
+    /// Appends an element to the back of the array-vector without spare capacity check.
+    ///
+    /// This method is useful when spare capacity check is already done by the caller.
+    ///
+    /// # Safety
+    ///
+    /// The caller must ensure that the array-vector has spare capacity to accommodate a new
+    /// element.
+    ///
+    /// # Example
+    /// ```rust
+    /// # use cds::array_vec;
+    /// let mut a = array_vec![3; usize; 1];
+    /// while a.has_spare_capacity() {
+    ///     unsafe { a.push_unchecked(0); }
+    /// }
+    /// assert_eq!(a, [1, 0, 0]);
+    /// ```
+    #[inline]
+    pub unsafe fn push_unchecked(&mut self, value: T) {
+        self.as_mut_ptr().add(self.len.as_usize()).write(value);
+        self.len += 1;
     }
 
     /// Removes the last element from an array-vector and returns it, or [`None`] if it is empty.
@@ -443,30 +512,6 @@ where
         e
     }
 
-    /// Appends an element to the back of the array-vector without spare capacity check.
-    ///
-    /// This method is useful when spare capacity check is already done by the caller.
-    ///
-    /// # Safety
-    ///
-    /// The caller must ensure that the array-vector has spare capacity to accommodate a new
-    /// element.
-    ///
-    /// # Example
-    /// ```rust
-    /// # use cds::array_vec;
-    /// let mut a = array_vec![3; usize; 1];
-    /// while a.has_spare_capacity() {
-    ///     unsafe { a.push_unchecked(0); }
-    /// }
-    /// assert_eq!(a, [1, 0, 0]);
-    /// ```
-    #[inline]
-    pub unsafe fn push_unchecked(&mut self, e: T) {
-        self.as_mut_ptr().add(self.len.as_usize()).write(e);
-        self.len += 1;
-    }
-
     /// Clears the array-vector, dropping all values.
     ///
     /// # Examples
@@ -652,7 +697,11 @@ where
     /// Tries to insert an element at position `index` within the vector, shifting all elements
     /// after it to the right.
     ///
-    /// Returns [`CapacityError`] if there is no spare capacity in the array vector.
+    /// Returns [`InsertError`] if there is no spare capacity in the array vector, or if `index` is
+    /// out of bounds.
+    ///
+    /// Note that in case of an error `value` is lost if it is not [`Copy`]. Use [`try_insert_val`]
+    /// to receive the element back in case of an error.
     ///
     /// Note that the worst case performance of this operation is O(n), because all elements of the
     /// array may be shifted right. If order of elements is not needed to be preserved,
@@ -660,32 +709,85 @@ where
     ///
     /// This is a non-panic version of [`insert`].
     ///
-    /// # Panics
+    /// # Examples
     ///
-    /// This method panics if `index > len()`.
+    /// ```rust
+    /// # use cds::{array_vec, errors::InsertError};
+    /// let mut v = array_vec![3; u64; 1, 2];
+    /// assert!(matches!(v.try_insert(3, 3), Err(InsertError::IndexOutOfBounds)));
+    ///
+    /// assert!(v.try_insert(0, 0).is_ok());
+    /// assert_eq!(v, [0, 1, 2]);
+    /// assert!(matches!(v.try_insert(1, 3), Err(InsertError::CapacityError)));
+    /// ```
+    ///
+    /// [`try_push`]: ArrayVec::try_push
+    /// [`try_insert_val`]: ArrayVec::try_insert_val
+    /// [`insert`]: ArrayVec::insert
+    #[inline]
+    pub fn try_insert(&mut self, index: usize, value: T) -> Result<(), InsertError> {
+        if index > self.len.as_usize() {
+            return Err(InsertError::IndexOutOfBounds);
+        }
+        if self.len >= Self::CAPACITY {
+            return Err(InsertError::CapacityError);
+        }
+        unsafe {
+            self.insert_unchecked(index, value);
+        }
+        Ok(())
+    }
+
+    /// Tries to insert an element at position `index` within the array-vector,
+    /// shifting all elements after it to the right.
+    ///
+    /// Returns [`InsertErrorVal`] if there is no spare capacity in the array vector,
+    /// or if `index` is out of bounds.
+    ///
+    /// The difference between this method and [`try_insert`], is that in case of an error
+    /// [`try_insert_val`] returns the element back to the caller.
+    ///
+    /// Note that the worst case performance of this operation is O(n), because all elements of the
+    /// array may be shifted right. If order of elements is not needed to be preserved,
+    /// use [`try_push_val`] instead.
+    ///
+    /// This is a non-panic version of [`insert`].
     ///
     /// # Examples
     ///
     /// ```rust
-    /// # use cds::{array_vec, errors::CapacityError};
+    /// # use cds::{array_vec, errors::InsertErrorVal};
     /// let mut v = array_vec![3; u64; 1, 2];
-    /// assert!(v.try_insert(0, 0).is_ok());
+    /// assert!(matches!(
+    ///     v.try_insert_val(5, 3),
+    ///     Err(InsertErrorVal::IndexOutOfBounds(v)) if v == 3
+    /// ));
+    /// assert_eq!(v, [1, 2]);
+    ///
+    /// assert!(v.try_insert_val(0, 0).is_ok());
+    /// assert_eq!(v, [0, 1, 2]);
+    ///
+    /// assert!(matches!(
+    ///     v.try_insert_val(1, 5),
+    ///     Err(InsertErrorVal::CapacityError(v)) if v == 5
+    /// ));
     /// assert_eq!(v, [0, 1, 2]);
-    /// assert!(matches!(v.try_insert(1, 3), Err(CapacityError)));
     /// ```
     ///
-    /// [`try_push`]: ArrayVec::try_push
+    /// [`try_insert_val`]: ArrayVec::try_insert_val
+    /// [`try_insert`]: ArrayVec::try_insert
+    /// [`try_push_val`]: ArrayVec::try_push_val
     /// [`insert`]: ArrayVec::insert
     #[inline]
-    pub fn try_insert(&mut self, index: usize, element: T) -> Result<(), CapacityError> {
-        if self.len >= Self::CAPACITY {
-            return Err(CapacityError {});
-        }
+    pub fn try_insert_val(&mut self, index: usize, value: T) -> Result<(), InsertErrorVal<T>> {
         if index > self.len.as_usize() {
-            panic!("index is out of bounds [0, {}]: {}", self.len, index);
+            return Err(InsertErrorVal::IndexOutOfBounds(value));
+        }
+        if self.len >= Self::CAPACITY {
+            return Err(InsertErrorVal::CapacityError(value));
         }
         unsafe {
-            self.insert_unchecked(index, element);
+            self.insert_unchecked(index, value);
         }
         Ok(())
     }
@@ -717,10 +819,10 @@ where
     /// unsafe { v.insert_unchecked(2, 2) };
     /// assert_eq!(v, [0, 1, 2]);
     /// ```
-    pub unsafe fn insert_unchecked(&mut self, index: usize, element: T) {
+    pub unsafe fn insert_unchecked(&mut self, index: usize, value: T) {
         let p = self.as_mut_ptr().add(index);
         ptr::copy(p, p.add(1), self.len.as_usize() - index);
-        p.write(element);
+        p.write(value);
         self.len += 1;
     }
 
